local version = 1

if version < towerofheaven.version then
	return -- Don't load the rest of the script.
end

-- Rules:
towerofheaven.rulesTable = {}
towerofheaven.rulesLogic = {}
towerofheaven.rulesValidIf = {}

towerofheaven.easyDifficulty = 1
towerofheaven.normalDifficulty = 2
towerofheaven.hardDifficulty = 3

local sneakerPanelSpecial = 6

local function isThrowableItem(item)
	if item == KITEM_BANANA or
		item == KITEM_EGGMAN or
		item == KITEM_ORBINAUT or
		item == KITEM_JAWZ or
		item == KITEM_MINE or
		item == KITEM_BALLHOG then
		return true
	end

	return false
end

local function isUsingDefaultThrowRules(player, item)
	local buttons = player.cmd.buttons
	local isPressingForward = (buttons & BT_FORWARD)
	local isPressingBackward = (buttons & BT_BACKWARD)
	
	if isPressingForward or isPressingBackward then
		return isPressingForward or isPressingBackward
	end
	
	-- Assuming it's a neutral state input.
	
	-- Banana rules.
	if item == KITEM_BANANA or item == KITEM_EGGMAN then
		return BT_BACKWARD
	-- Throwable rules.
	elseif item == KITEM_ORBINAUT or item == KITEM_JAWZ then
		return BT_FORWARD
	-- Mine rules.
	elseif item == KITEM_MINE then
		return BT_FORWARD
	-- Ballhog.
	elseif item == KITEM_BALLHOG then
		return BT_FORWARD
	end
	
	return 0
end

-- Do we have springs
local function doWeHaveSprings()
	for mobj in thinkers.iterate("mobj") do
		if mobj.flags & MF_SPRING then return true end
	end
	
	for sector in sectors.iterate do
		local springSpecial = GetSecSpecial(sector.special, 3)
		if springSpecial == 1 or springSpecial == 3 then return true end
	end
	
	return false
end

-- Do we have acro
local function doWeHaveAcro()
	return type(acrobatics) == "string" or acrobasics == true or OpenTricks_Table ~= nil
end

towerofheaven.getRulesTable = function(difficulty)
	local newRulesTable = {}
	
	for i = 1, #towerofheaven.rulesTable do
		if towerofheaven.rulesTable[i].difficulty == difficulty then
			table.insert(newRulesTable, towerofheaven.rulesTable[i])
		end
	end
	
	return newRulesTable
end

towerofheaven.getRuleByName = function(rulename)
	for i = 1, #towerofheaven.rulesTable do
		if towerofheaven.rulesTable[i].name == rulename then
			return towerofheaven.rulesTable[i]
		end
	end
	
	return nil
end

local function initRule(name, description, difficulty, codeToRun, params, validIf)
	local rule = {
		name = name,
		description = description,
		descriptionChosen = nil,
		difficulty = difficulty,
		params = params
	}
	
	table.insert(towerofheaven.rulesTable, rule)
	towerofheaven.rulesLogic[name] = codeToRun
	towerofheaven.rulesValidIf[name] = validIf
	
	return rule
end

towerofheaven.getRuleDescription = function(rule)
	return rule.descriptionChosen or rule.description
end

-- ...Let me just alias this for people who want to init custom rules.
towerofheaven.initRule = initRule

-- Easy rules:
local goldenBlockFlat = "TOHWALL3"
towerofheaven.goldenblockflat = goldenBlockFlat
towerofheaven.setGoldenBlockDefault = function()
	towerofheaven.goldenblockflat = goldenBlockFlat
end

table.insert(towerofheaven.ruleInits, function()
	if not mapheaderinfo[gamemap]["tohgoldenblockflat"] then
		towerofheaven.setGoldenBlockDefault()
		return
	end
	towerofheaven.goldenblockflat = mapheaderinfo[gamemap]["tohgoldenblockflat"]
end)

local noGoldenBlocks = initRule("goldenblocks", "Thou shalt not stand on\ngolden blocks.", towerofheaven.easyDifficulty, {["ThinkFrame"] = function(player, params)
	local mobj = player.mo
	
	-- Only care about the blocks we are touching.
	if not P_IsObjectOnGround(mobj) then return end
	
	-- Get the current sector...
	local sector = P_ThingOnSpecial3DFloor(mobj)
	if not sector then sector = mobj.subsector.sector end
	
	-- If flipped, we wanna get the ceiling's ceilingpic instead.
	local whichpic = "floorpic"
	if P_MobjFlip(mobj) == -1 then ceilingpic = "ceilingpic" end
	
	if sector[whichpic] == towerofheaven.goldenblockflat then towerofheaven.boltOfJudgement(player) end
end})

local noSneakerPanels = initRule("sneakerpanels", "Thou shalt not stand on\nSneaker Panels.", towerofheaven.easyDifficulty, {["ThinkFrame"] = function(player, params)
	local mobj = player.mo
	
	-- Only care about the blocks we are touching.
	if not P_IsObjectOnGround(mobj) then return end
	
	-- Get the current sector...
	local sector = P_ThingOnSpecial3DFloor(mobj)
	if not sector then sector = mobj.subsector.sector end
	
	if GetSecSpecial(sector.special, 4) == sneakerPanelSpecial then towerofheaven.boltOfJudgement(player) end
end})

local noOffRoadTimerLimit = TICRATE
local function offroadRestoreColor(mobj)
	if mobj.tohPrevColor then
		mobj.color = mobj.tohPrevColor
		mobj.tohPrevColor = nil
	end
end
local noOffRoad = initRule("offroad", "Thou shalt not stand on\noffroad for longer than a second.", towerofheaven.easyDifficulty, {["ThinkFrame"] = function(player, params)
	local mobj = player.mo
	
	offroadRestoreColor(mobj)
	
	-- Only care about the blocks we are touching.
	if not P_IsObjectOnGround(mobj) then return end
	
	-- Get the current sector...
	local sector = P_ThingOnSpecial3DFloor(mobj)
	if not sector then sector = mobj.subsector.sector end
	
	mobj.tohOffroadTimer = $ or 0
	
	-- Get offroad, run a timer up if the player's on offroad.
	local offroadSpecial = GetSecSpecial(sector.special, 1)
	if offroadSpecial >= 2 and offroadSpecial <= 4 then -- 2, 3, 4 for weak, normal and strong offroad.
		mobj.tohOffroadTimer = $ + 1
		if mobj.tohOffroadTimer % 2 == 0 then
			mobj.tohPrevColor = mobj.color
			mobj.color = SKINCOLOR_RED
		end
	elseif mobj.tohOffroadTimer > 0 then
		mobj.tohOffroadTimer = $ - 1
	end
	
	if mobj.tohOffroadTimer >= noOffRoadTimerLimit then
		mobj.tohOffroadTimer = 0
		offroadRestoreColor(mobj)
		towerofheaven.boltOfJudgement(player)
	end
end})

-- This emcompasses both spring PANELS and spring OBJECTS.
local function springCollision(player, params, targetMobj)
	-- If this target object is not a spring AND the target wasn't just sprung.
	if (targetMobj.flags & MF_SPRING) == MF_SPRING and (player.mo.eflags & MFE_SPRUNG) == 0 then
		towerofheaven.boltOfJudgement(player)
	end
end

local noSprings = initRule("springs", "Thou shalt not touch\nsprings.", towerofheaven.easyDifficulty,
{["ThinkFrame"] = function(player, params)
	local mobj = player.mo
	
	-- Only care about the blocks we are touching.
	if not P_IsObjectOnGround(mobj) then return end
	
	-- Get the current sector...
	local sector = P_ThingOnSpecial3DFloor(mobj)
	if not sector then sector = mobj.subsector.sector end
	
	-- Get spring sectors.
	local springSpecial = GetSecSpecial(sector.special, 3)
	if springSpecial == 1 or springSpecial == 3 then towerofheaven.boltOfJudgement(player) end
end,
["MobjMoveCollide"] = springCollision,
["MobjCollide"] = springCollision}, nil, do return doWeHaveSprings() end)

-- Medium rules:
local function collision(player, params, targetMobj)
	if not targetMobj.player then return end
	local targetPlayer = targetMobj.player
	if towerofheaven.isDamaged(targetPlayer) then return end
	
	towerofheaven.boltOfJudgement(player)
	if towerofheaven.isRuleActive("bonking", targetPlayer) then 
		towerofheaven.boltOfJudgement(targetPlayer)
	end
end

local noBonking = initRule("bonking", "Thou shalt not touch\nanother player.", towerofheaven.normalDifficulty,
{["MobjMoveCollide"] = collision, ["MobjCollide"] = collision}, nil, do return towerofheaven.pInGame() > 1 end)

local ROULETTE_ENDTIC = TICRATE*3
local noMashing = initRule("mashing", "Thou shalt not mash out\nyour item roulette.", towerofheaven.normalDifficulty, {["PreThinkFrame"] = function(player, params)
	local pks = player.kartstuff
	local roulettetic = pks[k_itemroulette]
		
	if roulettetic <= 0 then return end
	roulettetic = $ + 1 -- the function +1s the var inside, we simulate it.
	
	local roulettestop = TICRATE + (3*(towerofheaven.pInGame() - pks[k_position]))
	local canbemashed = roulettestop <= roulettetic

	if roulettetic <= ROULETTE_ENDTIC and canbemashed and (player.cmd.buttons & BT_ATTACK) == BT_ATTACK then towerofheaven.boltOfJudgement(player) end
end})

local noHonking = initRule("honking", "Thou shalt not honk\nyour horn.", towerofheaven.normalDifficulty, {["ThinkFrame"] = function(player, params)
	local hmstuff = player.hornmod
	
	-- Weirdly hacky but avoids me having to do all the checks that hornmod does.
	if hmstuff.horninput == 1 and hmstuff.flash >= 1 then towerofheaven.boltOfJudgement(player) end
end}, nil, do return type(HORNMOD) == "string" end) -- Only run this if hornmod exists.


local noBraking = initRule("braking", "Thou shalt not brake.", towerofheaven.normalDifficulty, {["ThinkFrame"] = function(player, params)
	local mobj = player.mo
	
	-- We can only brake on the ground.
	if not P_IsObjectOnGround(mobj) then return end
	if not (player.cmd.buttons & BT_BRAKE) then return end
	
	towerofheaven.boltOfJudgement(player)
end})

local noTricking = initRule("tricking", "Thou shalt not trick.", towerofheaven.normalDifficulty, {["ThinkFrame"] = function(player, params)
	if (player.trickcharge and player.trickcharge > 0) or -- Acrobatics
		(player.hastricked and player.hastricked == true) or -- Acrobasics
		(OpenTricks_TryingTrick and OpenTricks_TryingTrick(player)) then -- Open Tricks
		towerofheaven.boltOfJudgement(player)
	end
	
end}, nil, do return doWeHaveAcro() end) -- Only run this if hornmod exists.

-- Hard rules:
local left, right = 1, 2
local driftStrings = {"left", "right"}
local noDriftingThisDirection = initRule("drifting", "Thou shalt not drift\n%s.", towerofheaven.hardDifficulty, {["ThinkFrame"] = function(player, params)
	local pks = player.kartstuff
	local driftDir = pks[k_drift]
	if driftDir == 0 then return end
	
	if encoremode then driftDir = $ * -1 end
	
	if params.currentChoice == left and driftDir > 0 then
		towerofheaven.boltOfJudgement(player)
	elseif params.currentChoice == right and driftDir < 0 then
		towerofheaven.boltOfJudgement(player)
	end
end}, {choices = {left, right}, strings = driftStrings})

local forward, backward = BT_FORWARD, BT_BACKWARD
local throwStrings = {"forward", "backward"}
local noThrowingThisDirection = initRule("throwing", "Thou shalt not throw\nnor place items %s.", towerofheaven.hardDifficulty, {["PreThinkFrame"] = function(player, params)
	if (player.cmd.buttons & BT_ATTACK) == BT_ATTACK then
		player.tohHoldingAttack = $ or 0
		player.tohHoldingAttack = $ + 1
	else
		player.tohHoldingAttack = 0
	end
	
	local pks = player.kartstuff
	if not (pks[k_itemamount] > 0 and isThrowableItem(pks[k_itemtype]) or pks[k_eggmanheld] > 0) then return end
	
	local mobj = player.mo
	local currentlyOrbiting = mobj.hnext ~= nil
	if not currentlyOrbiting and pks[k_itemtype] ~= KITEM_BALLHOG then return end
	
	local cmd = player.cmd
	if player.tohHoldingAttack ~= 1 then return end
	
	-- Detect the default throw direction on a neutral input.
	local itemType = pks[k_itemtype]
	if itemType == 0 and pks[k_eggmanheld] > 0 then
		itemType = KITEM_EGGMAN
	end
	
	if isUsingDefaultThrowRules(player, itemType) == params.currentChoice or (player.cmd.buttons & params.currentChoice) then
		towerofheaven.boltOfJudgement(player)
	end
end}, {choices = {forward, backward}, strings = throwStrings}, do return towerofheaven.pInGame() > 1 end)

addHook("NetVars", function(network)
	if version < towerofheaven.version then return end
	towerofheaven.rulesTable = network($)
	towerofheaven.goldenblockflat = network($)
end)