local BEETLE_VERSION = 1

-- avoid redefiniton on updates
if beetlemobj == nil then rawset(_G, "beetlemobj", {}) end
if beetlemobj.BEETLE_VERSION == nil or beetlemobj.BEETLE_VERSION < BEETLE_VERSION then

freeslot("SPR_GRRL", "SPR_FRRL",
	"MT_DIGGLE_SKD_BEETLE", "sfx_beeto",
	"S_DIGGLE_SKD_BEETLEIDLE", "S_DIGGLE_SKD_BEETLEBONK")

mobjinfo[MT_DIGGLE_SKD_BEETLE] = {
	--$Name Shovel Knight Dig Bettle
	--$Sprite GRRLA0
	--$Category Diggle
	doomednum 	= 3902,
	spawnhealth = 9999,
	spawnstate 	= S_DIGGLE_SKD_BEETLEIDLE,
	seestate	= S_DIGGLE_SKD_BEETLEBONK,
	activesound	= sfx_beeto,
	speed 		= 20<<FRACBITS,
	radius 		= 64<<FRACBITS,
	height 		= 256<<FRACBITS,
	dispoffset	= 2,
	flags		= MF_SPECIAL|MF_SHOOTABLE|MF_BOUNCE|MF_ENEMY,
}

function A_Bettle_DetermineSprite(actor, var1, var2)
	if actor.extravalue2 == 1 then actor.sprite = SPR_FRRL
	else actor.sprite = SPR_GRRL
	end
end

states[S_DIGGLE_SKD_BEETLEIDLE] = {
	sprite = SPR_GRRL,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 3,
	var2 = 2,
	action = A_Bettle_DetermineSprite,
	nextstate = S_DIGGLE_SKD_BEETLEIDLE
}

states[S_DIGGLE_SKD_BEETLEBONK] = {
	sprite = SPR_GRRL,
	frame = E,
	tics = -1,
	action = A_Bettle_DetermineSprite,
	nextstate = S_DIGGLE_SKD_BEETLEBONK
}

addHook("MobjThinker", function(mo)
	if mo.info.spawnstate ~= nil and mo.state == mo.info.spawnstate then
		P_Thrust(mo, mo.angle, FixedMul(mo.info.speed / TICRATE, mapobjectscale))
		mo.extravalue1 = mo.angle
		mo.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
		
		local angleDiff = abs(AngleFixed(mo.angle) - AngleFixed(mo.extravalue1))
		
		if angleDiff > 45*FRACUNIT then
			mo.eflags = $ | MFE_JUSTBOUNCEDWALL
			local bump = P_SpawnMobj(mo.x, mo.y, mo.height/2, MT_BUMP)
			bump.scale = 3<<FRACBITS
			
			S_StartSound(mo, sfx_s3k49)
		
			P_SetObjectMomZ(mo, 10<<FRACBITS, true)
			mo.state = mo.info.seestate
			
			if angleDiff > 90*FRACUNIT then
				if mo.extravalue2 == 0 then mo.extravalue2 = 1
				else mo.extravalue2 = 0
				end
			end
		end
	elseif mo.info.seestate ~= nil and mo.state == mo.info.seestate then
		if P_IsObjectOnGround(mo) then
			mo.state = mo.info.spawnstate
		end
	end
	
end, MT_DIGGLE_SKD_BEETLE)

addHook("MobjCollide", function(thing, tmthing)
	-- Height checks.
	if tmthing.momz < 0 then
		if tmthing.z + tmthing.momz > thing.z + thing.height then
			return false
		end
	elseif tmthing.z > thing.z + thing.height then
		return false
	end
	
	if tmthing.momz > 0 then
		if tmthing.z + tmthing.height + tmthing.momz < thing.z then
			return false
		end
	elseif tmthing.z + tmthing.height < thing.z then
		return false
	end
	
	if tmthing.type == MT_ORBINAUT_SHIELD or
		tmthing.type == MT_JAWZ_SHIELD or
		tmthing.type == MT_BANANA_SHIELD or
		tmthing.type == MT_SSMINE or
		tmthing.type == MT_EGGMANITEM_SHIELD or
		tmthing.type == MT_SINK_SHIELD or 
		tmthing.type == MT_MINEEXPLOSION then
		return false
	elseif tmthing.type == MT_PLAYER then
		return nil
	end
	
	return true
	
end, MT_DIGGLE_SKD_BEETLE)

-- Prevent the mine from killing the beetle...

addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
	if inflictor and inflictor.valid then return true end
end, MT_DIGGLE_SKD_BEETLE)

addHook("MapLoad", function()
	-- *pained groan* "why does mobjspawn SUCK"
	for thing in mapthings.iterate do
		if thing.type ~= mobjinfo[MT_DIGGLE_SKD_BEETLE].doomednum then continue end
		if thing.options & MTF_OBJECTSPECIAL then
			thing.mobj.extravalue2 = 1
		end
	end
end)

addHook("MobjSpawn", function(mobj)
	P_SpawnShadowMobj(mobj)
end, MT_DIGGLE_SKD_BEETLE)

beetlemobj.BEETLE_VERSION = BEETLE_VERSION

end