local GNOME_VERSION = 7

-- avoid redefiniton on updates
if gnomemobj == nil then rawset(_G, "gnomemobj", {}) end
if gnomemobj.GNOME_VERSION == nil or gnomemobj.GNOME_VERSION < GNOME_VERSION then

-- Global moment
local FRACUNIT = FRACUNIT
local FRACBITS = FRACBITS
local TICRATE = TICRATE
local FF_ANIMATE = FF_ANIMATE

-- Variables
local noclipDistance = 1024<<FRACBITS
local moveDistance = 96<<FRACBITS
local standStillDistance = 64<<FRACBITS
local gnomeRespawnTime = 10*TICRATE
local hatTransferTime = TICRATE

-- Helpers
local function powerOf(num, to)
	local final = num
	for i = 1, to do final = FixedMul(final, num) end
	return final
end

local function easeOut(a)
	local frac = FRACUNIT - a
	return FRACUNIT - powerOf(frac, 2)
end

freeslot(
	-- sprites:
	"SPR_GNOM", "SPR_GNMW", "SPR_FNMW", "SPR_GNDT", "SPR_GNHT",
	-- objects:
	"MT_DIGGLE_GF_GNOME", "MT_DIGGLE_GF_GNOMESPAWN", "MT_DIGGLE_GF_RADIUS", "MT_DIGGLE_GF_ENDPOINT", "MT_DIGGLE_GF_HAT",
	-- normal states:
	"S_DIGGLE_GF_GNOMEIDLE", "S_DIGGLE_GF_GNOMEMOVE",
	-- death:
	"S_DIGGLE_GF_GNOMEDEAD1", "S_DIGGLE_GF_GNOMEDEAD2",
	"S_DIGGLE_GF_GNOMEDEAD3", "S_DIGGLE_GF_GNOMEDEAD4",
	"S_DIGGLE_GF_GNOMEDEAD5", "S_DIGGLE_GF_GNOMEDEAD6",
	"S_DIGGLE_GF_GNOMEDEAD7", "S_DIGGLE_GF_GNOMEDEAD8",
	"S_DIGGLE_GF_GNOMEDEAD9", "S_DIGGLE_GF_GNOMEDEAD10",
	-- hat states:
	"S_GNOMEHAT_INVISIBLE", "S_GNOMEHAT_FOLLOWME", "S_GNOMEHAT_DETACH",
	-- sounds:
	"sfx_gnomoh", "sfx_gnomf", "sfx_gnomgt", "sfx_gnomtr", "sfx_delivt", "sfx_mach3"
)

mobjinfo[MT_DIGGLE_GF_GNOME] = {
	--$Name Gnome Forest Gnome
	--$Sprite GNOMA0
	--$Category Diggle
	doomednum 	= 3903,
	spawnstate 	= S_DIGGLE_GF_GNOMEIDLE,
	seestate 	= S_DIGGLE_GF_GNOMEMOVE,
	deathstate  = S_DIGGLE_GF_GNOMEDEAD1,
	seesound 	= sfx_gnomgt,
	painsound	= sfx_gnomtr,
	attacksound = sfx_gnomoh,
	deathsound  = sfx_gnomf,
	radius 		= 64<<FRACBITS,
	height 		= 32<<FRACBITS,
	speed 		= 10<<FRACBITS,
	flags 		= MF_SPECIAL|MF_BOUNCE,
}

local deathTime = 10*TICRATE

states[S_DIGGLE_GF_GNOMEIDLE] = {
	sprite = SPR_GNOM,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 16,
	var2 = 2
}

states[S_DIGGLE_GF_GNOMEMOVE] = {
	sprite = SPR_GNMW,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 13,
	var2 = 2
}

states[S_DIGGLE_GF_GNOMEDEAD1] = {
	sprite = SPR_GNDT,
	frame = A,
	tics = 9,
	action = A_Scream,
	nextstate = S_DIGGLE_GF_GNOMEDEAD2
}

states[S_DIGGLE_GF_GNOMEDEAD2] = {
	sprite = SPR_GNDT,
	frame = B,
	tics = 9,
	nextstate = S_DIGGLE_GF_GNOMEDEAD3
}

states[S_DIGGLE_GF_GNOMEDEAD3] = {
	sprite = SPR_GNDT,
	frame = C,
	tics = 1,
	nextstate = S_DIGGLE_GF_GNOMEDEAD4
}

states[S_DIGGLE_GF_GNOMEDEAD4] = {
	sprite = SPR_GNDT,
	frame = D,
	tics = 1,
	nextstate = S_DIGGLE_GF_GNOMEDEAD5
}

states[S_DIGGLE_GF_GNOMEDEAD5] = {
	sprite = SPR_GNDT,
	frame = E,
	tics = 2,
	nextstate = S_DIGGLE_GF_GNOMEDEAD6
}

states[S_DIGGLE_GF_GNOMEDEAD6] = {
	sprite = SPR_GNDT,
	frame = F,
	tics = 1,
	nextstate = S_DIGGLE_GF_GNOMEDEAD7
}

states[S_DIGGLE_GF_GNOMEDEAD7] = {
	sprite = SPR_GNDT,
	frame = G,
	tics = 2,
	nextstate = S_DIGGLE_GF_GNOMEDEAD8
}

states[S_DIGGLE_GF_GNOMEDEAD8] = {
	sprite = SPR_GNDT,
	frame = H,
	tics = deathTime,
	nextstate = S_DIGGLE_GF_GNOMEDEAD9
}

states[S_DIGGLE_GF_GNOMEDEAD9] = {
	sprite = SPR_GNDT,
	frame = FF_TRANS40|H,
	tics = 5,
	nextstate = S_DIGGLE_GF_GNOMEDEAD10
}

states[S_DIGGLE_GF_GNOMEDEAD10] = {
	sprite = SPR_GNDT,
	frame = FF_TRANS80|H,
	tics = 5,
	nextstate = S_NULL
}

mobjinfo[MT_DIGGLE_GF_RADIUS] = {
	--$Name Gnome Forest Drop Off Point (w1028)
	--$Sprite GNOMHOS1
	--$Category Diggle
	doomednum 	= 3904,
	spawnstate 	= S_INVISIBLE,
	seesound	= sfx_delivt,
	radius 		= 1028<<FRACBITS,
	flags 		= MF_SPECIAL|MF_NOCLIPTHING,
}

mobjinfo[MT_DIGGLE_GF_ENDPOINT] = {
	--$Name Gnome Forest Gnome Position End Point
	--$Sprite GNOMSHR1
	--$Category Diggle
	doomednum 	= 3905,
	spawnstate 	= S_INVISIBLE,
}

mobjinfo[MT_DIGGLE_GF_GNOMESPAWN] = {
	spawnstate = S_INVISIBLE,
	flags = MF_NOBLOCKMAP|MF_NOSECTOR
}

states[S_GNOMEHAT_INVISIBLE] = {
	sprite = SPR_NULL,
	frame = A,
	tics = 1,
	nextstate = S_GNOMEHAT_FOLLOWME,
}

local function A_GF_GnomeHatFollower(hat, var1, var2)
	if hat.fuse > 0 then
		if hat.flags2 & MF2_DONTDRAW then hat.flags2 = $ & ~MF2_DONTDRAW end
			
		-- Hat flickers to indicate when it will despawn (also an indicator of boost timer).
		if leveltime % ((hat.fuse+TICRATE*2)/TICRATE) == 0 then
			hat.flags2 = $ | MF2_DONTDRAW
		end
	end

	if not (hat.target and hat.target.valid) then
		P_KillMobj(hat)
		return
	end
	
	local xOffset, yOffset = 0, 0
	if hat.extravalue1 > 0 then
		local xDiff = hat.x - hat.target.x
		local yDiff = hat.y - hat.target.y
			
		xOffset = $ + FixedMul(xDiff, FRACUNIT - easeOut((FRACUNIT / hatTransferTime) * (hatTransferTime - hat.extravalue1)))
		yOffset = $ + FixedMul(yDiff, FRACUNIT - easeOut((FRACUNIT / hatTransferTime) * (hatTransferTime - hat.extravalue1)))
		
		hat.extravalue1 = $ - 1
		K_SpawnSparkleTrail(hat)
	end
	
	P_MoveOrigin(hat, hat.target.x + xOffset, hat.target.y + yOffset, hat.target.z + hat.target.height)
end
states[S_GNOMEHAT_FOLLOWME] = {
	sprite = SPR_GNHT,
	frame = A,
	tics = 1,
	action = A_GF_GnomeHatFollower,
	nextstate = S_GNOMEHAT_FOLLOWME,
}

local function A_GF_GnomeHatDeath(mobj, var1, var2)
	if not (mobj.target and mobj.target.valid) then
		P_RemoveMobj(mobj)
		return
	end
	
	-- Killing the mobj applies this and I don't want it.
	if mobj.flags & MF_NOGRAVITY then mobj.flags = $ & ~MF_NOGRAVITY end
	
	mobj.momx = (mobj.target.momx/4)*3
	mobj.momy = (mobj.target.momy/4)*3
	mobj.momz = (mobj.target.momz/4)*3
	
	mobj.angle = mobj.target.angle + ANGLE_90
	P_Thrust(mobj, FixedAngle(P_RandomFixed() * 180) + mobj.angle + ANGLE_90, 16*mapobjectscale)
	P_SetObjectMomZ(mobj, 8*mapobjectscale, true)
	
	S_StartSoundAtVolume(mobj, sfx_s3k5d, 128)
	
	mobj.target = nil
end
states[S_GNOMEHAT_DETACH] = {
	sprite = SPR_GNHT,
	frame = A,
	tics = TICRATE*2,
	action = A_GF_GnomeHatDeath,
	nextstate = S_NULL,
}

mobjinfo[MT_DIGGLE_GF_HAT] = {
	spawnstate 	= S_GNOMEHAT_INVISIBLE,
	deathstate  = S_GNOMEHAT_DETACH,
	xdeathstate = S_GNOMEHAT_DETACH,
	radius		= 32*FRACUNIT,
	height		= 64*FRACUNIT,
	flags 		= MF_NOCLIPTHING,
}

local function getProjectileSpeed(mobjtype)
	if mobjtype == MT_BALLHOG then
		if gamespeed == 0 then
			return 68*mapobjectscale
		elseif gamespeed == 1 then
			return 82*mapobjectscale
		elseif gamespeed == 2 then
			return 96*mapobjectscale
		end
	else
		local projspeed = mobjinfo[mobjtype].speed
		if gamespeed == 0 then
			projspeed = FixedMul(projspeed, FRACUNIT - FRACUNIT/4) -- 75%
		elseif gamespeed == 2 then
			projspeed = FixedMul(projspeed, FRACUNIT + FRACUNIT/4) -- 125%
		end
		
		return FixedMul(projspeed, mapobjectscale)
	end
end

local function isPlayerDamaged(player)
	return --[[player.powers[pw_flashing] > 0 or ]] player.kartstuff[k_squishedtimer] > 0 or player.kartstuff[k_spinouttimer] > 0 or
			--[[player.kartstuff[k_invincibilitytimer] > 0 or player.kartstuff[k_growshrinktimer] > 0 or player.kartstuff[k_hyudorotimer] > 0 or]]
			(gametype == GT_MATCH and ((player.kartstuff[k_bumper] <= 0 and player.kartstuff[k_comebacktimer]) or player.kartstuff[k_comebackmode] == 1))

end

-- Gnome Logic
local dropGnome = nil
local swapGnome = nil
local setUpHat = nil

local function gnomeGetEffects(gnome)
	if not (gnome.target and gnome.target.valid) then return end
	
	local dist = FixedHypot(gnome.x - gnome.target.x, gnome.y - gnome.target.y)
	local randomness = P_RandomKey(8) -- 1/8
	local soundToPlay = gnome.info.seesound
	
	if dist >= noclipDistance then
		soundToPlay = gnome.info.painsound
		gnome.extravalue2 = 1
	end
	
	if randomness == 0 then
		S_StartSound(gnome.target, gnome.info.attacksound)
	end
	
	S_StartSound(gnome.target, soundToPlay)
	gnome.extravalue1 = TICRATE
	
	gnome.color = gnome.target.player.skincolor
	if not gnome.tracer then gnome.tracer = P_SpawnMobj(gnome.x, gnome.y, gnome.z, MT_DIGGLE_GF_HAT) end
	setUpHat(gnome)
end

dropGnome = function(haver, toPrevious)
	if toPrevious then
		-- Give the gnome to the previous placer.
		-- LOST FOREVER otherwise if no previous player.
		local pks = haver.player.kartstuff
		local position = pks[k_position]
		
		local closestPosition = #players
		local playerFound = nil
		
		for otherPlayer in players.iterate do
			if otherPlayer.spectator or not (otherPlayer.mo and otherPlayer.mo.valid) then continue end
			if haver == otherPlayer.mo then continue end -- ignore yourself.
			-- Don't pass the gnome to this player if they're dead.
			if otherPlayer.playerstate == PST_DEAD then continue end
			
			local otherPosition = otherPlayer.kartstuff[k_position]
			if otherPosition > position and otherPosition < closestPosition then
				closestPosition = otherPosition
				playerFound = otherPlayer
			end
		end
	
		if playerFound then
			swapGnome(haver, playerFound.mo)
		else
			haver.gnome.target = nil
			P_KillMobj(haver.gnome) -- o7
			haver.gnome = nil
		end
	else
		-- Respawn the gnome after the time defined here.
		haver.gnome.lastlook = gnomeRespawnTime
		
		haver.gnome.color = SKINCOLOR_CREAM
		haver.gnome.target = nil
		--[[ if haver.gnome.tracer then
			P_KillMobj(haver.gnome.tracer)
		end]]
		haver.gnome = nil
	end
	
	return true
end

swapGnome = function(haver, taker)
	-- Both have gnomes, don't swap.
	if taker.gnome and haver.gnome then
		-- Instead drop the haver's gnome.
		dropGnome(haver)
		return false
	end

	taker.gnome = haver.gnome
	taker.gnome.target = taker
	haver.gnome = nil
	
	gnomeGetEffects(taker.gnome)
	
	return true
end

setUpHat = function(gnome)
	local hat = gnome.tracer
	hat.color = gnome.color
	if hat.target and hat.target.valid then -- Hat transfer animation stuff.
		hat.extravalue1 = hatTransferTime
	end
	hat.target = gnome.target
end

addHook("TouchSpecial", function(gnome, playermobj)
	if playermobj.type ~= MT_PLAYER then return true end -- Only players can get the Gnome
	if gnome.target and gnome.target.valid then return true end -- Gnome mustn't have a target already.
	if playermobj.gnome and playermobj.gnome.valid then return true end -- The player can't be escorting another gnome (that'd be cheating!)
	if gnome.health <= 0 then return true end -- Gnome must be alive.
	-- You're not able to pick the gnome if you just got hurt, are hurt, etc.
	if isPlayerDamaged(playermobj.player) then return true end
	
	-- Just in case, run this here too in case the players run into the gnome on a frame
	-- where he is visually gone.
	if gnome.flags2 & MF2_DONTDRAW then gnome.flags2 = $ & ~MF2_DONTDRAW end

	gnome.target = playermobj
	playermobj.gnome = gnome
	gnomeGetEffects(gnome)
	
	return true
end, MT_DIGGLE_GF_GNOME)

addHook("MobjThinker", function(mobj)
	if mobj.health <= 0 then return end

	if mobj.extravalue1 > 0 then
		K_SpawnSparkleTrail(mobj)
		mobj.extravalue1 = $ - 1
	end

	if not (mobj.target and mobj.target.valid) then
		if mobj.state ~= S_DIGGLE_GF_GNOMEIDLE then
			mobj.state = S_DIGGLE_GF_GNOMEIDLE
			mobj.friction = 28*FRACUNIT/32
		end
		
		if mobj.tracer and mobj.tracer.valid then
			P_KillMobj(mobj.tracer)
			mobj.tracer = nil
		end
		
		if mobj.lastlook > 0 then
			if mobj.flags2 & MF2_DONTDRAW then mobj.flags2 = $ & ~MF2_DONTDRAW end
			
			-- Gnome flickers to indicate when it will despawn.
			if leveltime % ((mobj.lastlook+TICRATE*2)/TICRATE) == 0 then
				mobj.flags2 = $ | MF2_DONTDRAW
			end
			
			mobj.lastlook = $ - 1
			-- Respawn the gnome - we destroy it rather than killing it.
			if mobj.lastlook == 0 then
				P_RemoveMobj(mobj)
			end
		end
		
		return
	end
	
	-- P_MoveOrigin(mobj, mobj.target.x, mobj.target.y, mobj.target.z)
	local following = mobj.target
	local dist = FixedHypot(mobj.x - following.x, mobj.y - following.y)
	local totalSpeed = abs(mobj.momx + mobj.momy + mobj.momz) / 3
	mobj.angle = R_PointToAngle2(mobj.x, mobj.y, following.x, following.y) -- Always look at your target.
	
	local finalNoclipDistance = noclipDistance
	if mobj.target.type == MT_DIGGLE_GF_ENDPOINT then
		finalNoclipDistance = $ / 4
	end
	
	-- Wow! The target's too far away. Let's get you noclip so you can catch up.
	if dist > finalNoclipDistance and not (mobj.flags & MF_NOCLIP) then
		mobj.flags = $ | MF_NOCLIP
	elseif dist <= moveDistance and mobj.flags & MF_NOCLIP then
		mobj.flags = $ & ~MF_NOCLIP
		
		if mobj.extravalue2 > 0 then
			mobj.extravalue1 = TICRATE
			mobj.extravalue2 = 0
			
			S_StartSound(mobj.target, mobj.info.seesound)
		end
	end
	
	if mobj.flags & MF_NOCLIP or not P_IsObjectOnGround(mobj.target) then
		local zDist = mobj.z - following.z
		local strength = mobj.info.speed
		
		if zDist > 0 then strength = -$ end
		P_SetObjectMomZ(mobj, strength, false)
	end
	
	if dist > moveDistance then
		if mobj.state ~= S_DIGGLE_GF_GNOMEMOVE then
			mobj.state = S_DIGGLE_GF_GNOMEMOVE
		end
		
		if following.angle - mobj.angle > 0 then
			mobj.sprite = SPR_GNMW
		else
			mobj.sprite = SPR_FNMW
		end
		
		mobj.friction = 29*FRACUNIT/32
		
		local finalSpeed = mobj.info.speed
		local lowBound, highBound = moveDistance, FixedMul(moveDistance, 130*(FRACUNIT/100))
		
		local ourDist = dist
		local multiplier = FixedDiv(ourDist - lowBound, highBound - lowBound)
		-- print(multiplier/(FRACUNIT/100) + "%")
		
		P_InstaThrust(mobj, mobj.angle, FixedMul(finalSpeed, FRACUNIT + multiplier))
	elseif dist <= standStillDistance or totalSpeed < mobj.info.speed then
		if mobj.state ~= S_DIGGLE_GF_GNOMEIDLE then
			mobj.state = S_DIGGLE_GF_GNOMEIDLE
		end
		
		mobj.friction = 28*FRACUNIT/32
	end	
end, MT_DIGGLE_GF_GNOME)

addHook("MobjCollide", function(thing, tmthing)
	if leveltime < 3 then return nil end
	
	-- Height checks.
	if tmthing.momz < 0 then
		if tmthing.z + tmthing.momz > thing.z + thing.height then
			return false
		end
	elseif tmthing.z > thing.z + thing.height then
		return false
	end
	
	if tmthing.momz > 0 then
		if tmthing.z + tmthing.height + tmthing.momz < thing.z then
			return false
		end
	elseif tmthing.z + tmthing.height < thing.z then
		return false
	end
	
	if tmthing.type ~= MT_PLAYER then return nil end
	
	if thing.player.kartstuff[k_squishedtimer] or thing.player.kartstuff[k_hyudorotimer] or
		thing.player.kartstuff[k_justbumped] or
		tmthing.player.kartstuff[k_squishedtimer] or tmthing.player.kartstuff[k_hyudorotimer] or
		tmthing.player.kartstuff[k_justbumped]  then
		return nil
	end
	
	if thing.scale > tmthing.scale + (mapobjectscale/8) or tmthing.scale > thing.scale + (mapobjectscale/8) then
		return nil
	end

	if gametype == GT_MATCH and
		((thing.player.kartstuff[k_bumper] and not tmthing.player.kartstuff[k_bumper]) or
		(tmthing.player.kartstuff[k_bumper] and not thing.player.kartstuff[k_bumper])) then
		return nil
	end
	
	if thing.gnome or tmthing.gnome then
		local gnomeHaver, nonGnomeHaver
		if thing.gnome and tmthing.gnome then
			-- Both players have gnomes - don't swap.
			return nil
		elseif thing.gnome then
			gnomeHaver = thing
			nonGnomeHaver = tmthing
		else
			gnomeHaver = tmthing
			nonGnomeHaver = thing
		end
		
		if gnomeHaver.health <= 0 then return nil end

		-- You're not able to pick the gnome if you just got hurt, are hurt, etc.
		if isPlayerDamaged(nonGnomeHaver.player) then return nil end
		
		swapGnome(gnomeHaver, nonGnomeHaver)
	end
	
	return nil

end, MT_PLAYER)

local function damageHandler(player, inflictor, source)
	local target = player.mo
	if not (target and target.gnome and target.gnome.valid) then return end
	local gnome = target.gnome
	if gnome.health <= 0 then return end

	if source ~= nil and inflictor ~= nil and target ~= source and
		source.type == MT_PLAYER then
		-- Sourced damage results in the gnome being transferred
		-- or dropped depending on circumstances.
		
		-- Swapped
		-- Shield hits and snipes are unambiguous full contact and direct hits.
		if inflictor.type == MT_ORBINAUT_SHIELD or
		inflictor.type == MT_JAWZ_SHIELD or
		inflictor.type == MT_BANANA_SHIELD or
		-- inflictor.type == MT_EGGMANITEM_SHIELD or
		(inflictor.type == MT_SSMINE or inflictor.type == MT_SSMINE_SHIELD and
		(inflictor.previousstate and inflictor.previousstate >= S_SSMINE_DEPLOY8 and inflictor.previousstate <= S_SSMINE_DEPLOY13)) or
		(inflictor.type == MT_BANANA and inflictor.health > 1) or
		-- (inflictor.type == MT_EGGMANITEM and inflictor.health > 1) or -- See the TouchSpecial hook just below.
		inflictor.type == MT_BALLHOG or
		inflictor.type == MT_SINK or
		-- Throwables like these need a speed check due to dropped items being the same type.
		((inflictor.type == MT_JAWZ or inflictor.type == MT_JAWZ_DUD) and inflictor.movefactor > FRACUNIT) or
		(inflictor.type == MT_ORBINAUT and
		(inflictor.movefactor > getProjectileSpeed(inflictor.type)/8 or
		(inflictor.movefactor <= getProjectileSpeed(inflictor.type)/8 and inflictor.health >= 7 and
		abs(inflictor.momx + inflictor.momy + inflictor.momz)/3 > FRACUNIT))) or
		inflictor == source -- Thunder Shield (?)
		then
			swapGnome(target, source)
		else
			dropGnome(target)
		end
	else
	-- Sourceless damage results in the gnome being dropped.
		dropGnome(target)
	end
end

addHook("PlayerSpin", damageHandler, MT_PLAYER)
addHook("PlayerExplode", damageHandler, MT_PLAYER)
addHook("PlayerSquish", function(player, inflictor, source)
	local target = player.mo
	if not (target and target.valid and target.gnome and target.gnome.valid) then return end
	local gnome = target.gnome
	if gnome.health <= 0 then return end

	if source ~= nil and inflictor ~= nil and target ~= source and
		source.type == MT_PLAYER then
		swapGnome(target, source)
	else
		dropGnome(target)
	end
end)

local function eggmanStealHandler(special, toucher)
	if not (special.target and special.target.valid) then return nil end
	if not (toucher.gnome and toucher.gnome.valid) then return nil end
	local itemOwner = special.target
	local gnomeOwner = toucher
	if gnomeOwner == itemOwner then return nil end
	if toucher.gnome.health <= 0 then return end
	
	-- Simply touching a dragged eggbox swaps the gnome.
	-- However on throw, only eggbox snipes swap the gnome.
	if special.type == MT_EGGMANITEM and special.health > 1 then 
		S_StartSound(toucher, sfx_bsnipe)
		swapGnome(gnomeOwner, itemOwner)
	elseif special.type == MT_EGGMANITEM_SHIELD then
		swapGnome(gnomeOwner, itemOwner)
	end
	
	return nil
end

addHook("TouchSpecial", eggmanStealHandler, MT_EGGMANITEM_SHIELD)
addHook("TouchSpecial", eggmanStealHandler, MT_EGGMANITEM)

local function minePreviousState(mobj)
	mobj.previousstate = mobj.state
end

addHook("MobjThinker", minePreviousState, MT_SSMINE)
addHook("MobjThinker", minePreviousState, MT_SSMINE_SHIELD)

addHook("MobjDeath", function(target, inflictor, source)
	if not (target.gnome and target.gnome.valid) then return end
	local gnome = target.gnome
	if gnome.health <= 0 then return end
	
	if inflictor and inflictor.type == MT_SINK and inflictor.target ~= nil then
		swapGnome(target, inflictor.target)
	else
		dropGnome(target, true)
	end
	
end, MT_PLAYER)

addHook("MobjSpawn", function(mobj)
	mobj.color = SKINCOLOR_CREAM
end, MT_DIGGLE_GF_GNOME)

addHook("MobjSpawn", function(mobj)
	mobj.health = 9999
	mobj.color = SKINCOLOR_CREAM
end, MT_DIGGLE_GF_HAT)

addHook("MobjDeath", function(mobj)
	mobj.color = SKINCOLOR_CREAM
	if mobj.tracer and mobj.tracer.valid then P_KillMobj(mobj.tracer) end -- Remove the hat.
end, MT_DIGGLE_GF_GNOME)

addHook("MobjFuse", function(mobj)
	if mobj.tracer and mobj.tracer.valid then P_RemoveMobj(mobj.tracer) end -- Remove the hat.
	P_RemoveMobj(mobj)
end, MT_DIGGLE_GF_GNOME)

addHook("MobjThinker", function(mobj)
	if mobj.state ~= S_GNOMEHAT_DETACH then return end
	
	if mobj.flags2 & MF2_DONTDRAW then mobj.flags2 = $ & ~MF2_DONTDRAW end
	if leveltime % 2 == 0 then mobj.flags2 = $ | MF2_DONTDRAW end
	-- print("Death timer: " + mobj.tics)
end, MT_DIGGLE_GF_HAT)

-- Radius Logic
local gnomeBoost = "juggy_mach3"
local speedBuff = 50*(FRACUNIT/100)
local accelBuff = 50*(FRACUNIT/100)

addHook("MobjThinker", function(mobj)
	-- We only care about Gnome-holders.
	searchBlockmap("objects", function(refmobj, playermobj)
		if leveltime == 3 and playermobj.type == MT_DIGGLE_GF_ENDPOINT then
			-- I'm assuming you won't place more than one, that'd be silly.
			refmobj.tracer = playermobj
			return true
		elseif leveltime == 4 and refmobj.tracer == nil then
			refmobj.tracer = refmobj
		end
		
		if FixedHypot(FixedHypot(refmobj.x - playermobj.x, refmobj.y - playermobj.y),
					refmobj.z - playermobj.z) > refmobj.info.radius then return nil end -- In radius?
		if not (playermobj.gnome and playermobj.gnome.valid) then return nil end -- Has gnome?
		if not P_IsObjectOnGround(playermobj) then return end

		local gnome = playermobj.gnome
		
		local gnomeBoostTimer = (3 + (playermobj.player.kartstuff[k_position] - 1))*TICRATE
		gnome.fuse = deathTime
		gnome.target = refmobj.tracer
		gnome.tracer.fuse = gnomeBoostTimer
		gnome.tracer = nil -- Don't hold the hat hostage, it is an independant entity now.
		playermobj.gnome = nil
		-- Three seconds base of boost + 1 second per position behind 1st.
		playermobj.gnomeboost = gnomeBoostTimer
		
		S_StartSound(playermobj, refmobj.info.seesound)
		P_LinedefExecute(mobj.extravalue2)
		return true
	end, mobj, 	mobj.x - mobj.info.radius/2, mobj.x + mobj.info.radius/2,
				mobj.y - mobj.info.radius/2, mobj.y + mobj.info.radius/2)
end, MT_DIGGLE_GF_RADIUS)

addHook("MobjThinker", function(mo)
	if not mo.gnomeboost then return end
	local player = mo.player

	if BLib then
		BLib.SetBoost(player, gnomeBoost, speedBuff, accelBuff, nil, true, false, false)
	else
		player.kartstuff[k_speedboost] = max($, speedBuff)
		player.kartstuff[k_accelboost] = max($, accelBuff)
	end

	local volume = 0
	
	local lowSpeedBound, highSpeedBound = 24<<FRACBITS, 48<<FRACBITS
	if gamespeed == 0 then
		lowSpeedBound = FixedMul($, FRACUNIT - FRACUNIT/4)
		highSpeedBound = FixedMul($, FRACUNIT - FRACUNIT/4)
	elseif gamespeed == 2 then
		lowSpeedBound = FixedMul($, FRACUNIT + FRACUNIT/4)
		highSpeedBound = FixedMul($, FRACUNIT + FRACUNIT/4)
	end
	lowSpeedBound = FixedMul($, mapobjectscale)
	highSpeedBound = FixedMul($, mapobjectscale)
	
	local multiplier = FixedDiv(player.speed - lowSpeedBound, highSpeedBound - lowSpeedBound)
	volume = FixedInt(FixedMul(255<<FRACBITS, multiplier))

	-- print(volume)
	-- Can't exactly dynamically change volume but I may as well use the system to determine if it should play.
	if volume > 128 then
		if not S_SoundPlaying(mo, sfx_mach3) then S_StartSound(mo, sfx_mach3) end
	else
		S_StopSoundByID(mo, sfx_mach3)
	end
	
	if leveltime % 6 == 0 and abs(mo.momx + mo.momy + mo.momz) / 3 > mo.scale then
		local ghost = P_SpawnGhostMobj(mo)
		ghost.colorized = true
		if leveltime % 12 == 0 then
			ghost.color = SKINCOLOR_RED
		else
			ghost.color = SKINCOLOR_YELLOW
		end
	end
	
	mo.gnomeboost = $ - 1
	if mo.gnomeboost == 0 then
		S_StopSoundByID(mo, sfx_mach3)
	
		if BLib then
			BLib.SetBoost(player, gnomeBoost)
		else
			-- player.kartstuff[k_speedboost] = 0
			-- player.kartstuff[k_accelboost] = 0
		end
	end
end, MT_PLAYER)

addHook("MapLoad", function()
	for thing in mapthings.iterate do
		if not thing.mobj then continue end
	
		if thing.mobj.type == MT_DIGGLE_GF_RADIUS then
			thing.mobj.extravalue2 = thing.angle
			continue
		end
		
		if thing.mobj.type == MT_DIGGLE_GF_GNOME then
			local gnomeSpawn = P_SpawnMobj(thing.mobj.x, thing.mobj.y, thing.mobj.z, MT_DIGGLE_GF_GNOMESPAWN)
			gnomeSpawn.target = thing.mobj
			continue
		end
	end
end)

-- Spawnpoint Logic
addHook("MobjThinker", function(mobj)
	if mobj.target and mobj.target.valid then return end
	-- If we reach this point our gnome is dead. Respawn him.
	
	mobj.target = P_SpawnMobj(mobj.x, mobj.y, mobj.z, MT_DIGGLE_GF_GNOME)
end, MT_DIGGLE_GF_GNOMESPAWN)

gnomemobj.GNOME_VERSION = GNOME_VERSION

end